#include "MouseInteractorStylePP.h"
#include "const.h"
#include "utils.h"
#include "parallel.h"
#include <sys/time.h>

MouseInteractorStylePP::MouseInteractorStylePP()
{
	this->clicked = FALSE;
}

/*! 
 * \brief Setter of private attribute pdcage
 * \public \memberof MouseInteractorStylePP 
 * 
 * This function is called internally by ::visualize
 *
 * \param[in] pd	vtkPolyData object of cage
 *
*/
void MouseInteractorStylePP::SetPolyDataCage (vtkSmartPointer<vtkPolyData> pd)
{
	this->pdcage = pd;
}

void MouseInteractorStylePP::SetPolyDataFigure (vtkSmartPointer<vtkPolyData> pd)
{
	this->pdfigure = pd;
}

void MouseInteractorStylePP::SetCageRef (vtkSmartPointer<vtkActor> ref)
{
	this->cage_ref = ref;
}

void MouseInteractorStylePP::SetMeanValueCoords (float *v)
{
	this->mvcoords = v;
}

void MouseInteractorStylePP::SetClData (struct cl_data *d)
{
	this->clData = d;
}

void MouseInteractorStylePP::OnMiddleButtonDown() 
{
	double picked[3];
	vtkSmartPointer<vtkPointPicker> pp;
	
	this->clicked = TRUE;

	// Get picked point id
	std::cout << "Picking pixel: " << this->Interactor->GetEventPosition()[0] << " " << this->Interactor->GetEventPosition()[1] << std::endl;
	this->Interactor->GetPicker()->Pick(this->Interactor->GetEventPosition()[0], this->Interactor->GetEventPosition()[1], 0, /* always zero.*/ this->GetCurrentRenderer());
	this->Interactor->GetPicker()->GetPickPosition(picked);
	std::cout << "Picked value down: " << picked[0] << " " << picked[1] << " " << picked[2] << std::endl;
	pp =  vtkPointPicker::SafeDownCast(this->Interactor->GetPicker());
	this->picked_point_down_id = pp->GetPointId();
	std::cout << "Picked point down id: " << this->picked_point_down_id << std::endl;
	this->picked_actor_down = pp->GetActor();
	std::cout << "Picked actor: " << this->picked_actor_down << std::endl;

	// Forward events
	//vtkInteractorStyleTrackballCamera::OnMiddleButtonDown();
}

void MouseInteractorStylePP::OnMiddleButtonUp()
{
	this->clicked = FALSE;
	
	// Forward events
	//vtkInteractorStyleTrackballCamera::OnMiddleButtonUp();
}

void MouseInteractorStylePP::OnMouseMove()
{
	double picked_point_up[3];
	unsigned int fig_npts, cage_npts;

	if (this->clicked)
	{
		// Get picked point coords
		this->Interactor->GetPicker()->Pick(this->Interactor->GetEventPosition()[0], this->Interactor->GetEventPosition()[1], 0, /* always zero.*/ this->GetCurrentRenderer());
		this->Interactor->GetPicker()->GetPickPosition(picked_point_up);
		std::cout << "Picked value up: " << picked_point_up[0] << " " << picked_point_up[1] << " " << picked_point_up[2] << std::endl;

		// Update x,y,z coords of moved point from Cage
		if (this->picked_point_down_id >= 0 && this->picked_actor_down == this->cage_ref)	// Check if picked point in down action is a valid vertex and if picked the cage
		{
			// Update point of cage
			this->pdcage->GetPoints()->SetPoint(picked_point_down_id, picked_point_up);
			this->pdcage->Modified();
			
			fig_npts = this->pdfigure->GetNumberOfPoints();
			cage_npts = this->pdcage->GetNumberOfPoints();				

			struct timeval start, end;
			long mtime, seconds, useconds;		

			#ifndef OPENCL
				double p[3], pc[3];
				unsigned int i, k;

				gettimeofday(&start, NULL);

				// Update every point of figure by using mean value coordinates
				for (k=0; k<fig_npts; k++)
				{
					p[0] = 0.0;
					p[1] = 0.0;
					p[2] = 0.0;
					for (i=0; i<cage_npts; i++)
					{
						// Recover point with mean value coordinates
						this->pdcage->GetPoints()->GetPoint(i, pc);
						p[0] += mvcoords[k*cage_npts+i] * pc[0];
						p[1] += mvcoords[k*cage_npts+i] * pc[1];
						p[2] += mvcoords[k*cage_npts+i] * pc[2];   	
					}
					// Update k-essim point
					this->pdfigure->GetPoints()->SetPoint(k, p);	
				}

				gettimeofday(&end, NULL);
				seconds  = end.tv_sec  - start.tv_sec;
				useconds = end.tv_usec - start.tv_usec;

				mtime = ((seconds) * 1000 + useconds/1000.0) + 0.5;
				printf("Elapsed time RP: %ld milliseconds\n", mtime);

			#else
				float* cage_points = reinterpret_cast<float*>(pdcage->GetPoints()->GetVoidPointer(0));
				float* figure_points = reinterpret_cast<float*>(pdfigure->GetPoints()->GetVoidPointer(0));

				gettimeofday(&start, NULL);

				recover_parallel_point(this->clData, cage_points, cage_npts, fig_npts, figure_points);

				gettimeofday(&end, NULL);
				seconds  = end.tv_sec  - start.tv_sec;
				useconds = end.tv_usec - start.tv_usec;

				mtime = ((seconds) * 1000 + useconds/1000.0) + 0.5;
				printf("Elapsed time RP: %ld milliseconds\n", mtime);

			#endif

			this->pdfigure->Modified(); 						
		
			// Refresh de view
			this->GetCurrentRenderer()->GetRenderWindow()->Render();	
		}
	 }
		
	// Forward events
	vtkInteractorStyleTrackballCamera::OnMouseMove();
}

vtkStandardNewMacro(MouseInteractorStylePP);
