//#pragma OPENCL EXTENSION cl_khr_fp64 : enable

#define TRUE 1
#define FALSE 0

__kernel void recover_point (__global float *mvcoords, __global float *cage, unsigned int cage_npts, __global float *out)
{
	float out_x = 0.0f, out_y = 0.0f, out_z = 0.0f;
	unsigned int i;

	// Get global index in mvcoords array
	unsigned int idx = get_global_id(0);
	unsigned int k = idx*cage_npts;

	// Calculate recovered point p
	for (i=0; i<cage_npts; i++)
	{
		out_x += mvcoords[k+i] * cage[i*3+0];
		out_y += mvcoords[k+i] * cage[i*3+1];
		out_z += mvcoords[k+i] * cage[i*3+2];
	}

	// Store point p in out 	 
	out[idx*3+0] = out_x;
	out[idx*3+1] = out_y;
	out[idx*3+2] = out_z;
}

__kernel void compute_coordinates (unsigned int cage_npts, __global float *cage_points, unsigned int nfaces, __global unsigned int *faces, __global float *x, __global float *w, __global float *dist, __global float *u)
{
	int sign;
	unsigned int nt=nfaces, npc=cage_npts, i, k, next, prev, p, inext, iprev, jump=FALSE;
	float l[3], theta[3], h, c[3], s[3], res, sum=0.0f, wi, ll, TOL=0.00001f;
	unsigned int idx = get_global_id(0);
	
	// Per cada vertex de la caixa calculem el modul de la distancia entre cada punt de la caixa i el punt x
	for (i=0; i<npc; i++)
	{
		dist[idx*npc+i] = sqrt(pow(cage_points[i*3+0]-x[idx*3+0], 2.0f) + pow(cage_points[i*3+1]-x[idx*3+1], 2.0f) + pow(cage_points[i*3+2]-x[idx*3+2], 2.0f));	// Calcul Modul	

		if (dist[idx*npc+i] < TOL)	// Si el punt es molt a prop de la caixa assignem 1 al pes corresponent i la resta 0's 
		{		
			w[idx*npc+i] = 1.0f;
			return;
		}
		
		// Normalitzem el vector distancia pel seu modul
		u[idx*npc*3+i*3+0] = (cage_points[i*3+0]-x[idx*3+0]) / dist[idx*npc+i];
		u[idx*npc*3+i*3+1] = (cage_points[i*3+1]-x[idx*3+1]) / dist[idx*npc+i];
		u[idx*npc*3+i*3+2] = (cage_points[i*3+2]-x[idx*3+2]) / dist[idx*npc+i];
	}

	for (k=0; k<nt; k++)	// Per cada triangle de la caixa
	{		
		h = 0.0f;
		for (i=0; i<3; i++)	//Per cada vertex de cada triangle
		{		 	
			// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles
			next = (i+1+3) % 3;
			prev = (i-1+3) % 3;

			// Calculem el modul de la distancia entre el punt seguent menys l'anterior
			ll = 0.0f;
			for (p=0; p<3; p++)
			{
				inext = faces[k*3+next];
				iprev = faces[k*3+prev];
				ll += pow(u[idx*npc*3+inext*3+p] - u[idx*npc*3+iprev*3+p], 2.0f);
			}

			l[i] = sqrt(ll);

			// Calculem l'angle theta
			theta[i] = 2.0f * asin((l[i]/2.0f));
			
			//Calculem h
			h += theta[i] / 2.0f; 
		}
		
		// Preguntem si el punt x es sobre el triangle
		if ((M_PI - h) < TOL)
		{			
			// Netejem els pesos
			for (i=0; i<npc; i++)
				w[idx*npc+i] = 0.0f;

			// Netejem el sumatori
			sum = 0.0f;

			// Anem a calcular els pesos wi
			for (i=0; i<3; i++)
			{
				// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles
				next = (i+1+3) % 3;
				prev = (i-1+3) % 3;

				// Computem els pesos
				wi = sin(theta[i]) * l[prev] * l[next];
				w[idx*npc+faces[k*3+i]] += wi; 			
				sum += wi;
			}
		
			// Sortim de l'evaluacio dels triangles, anem a normalitzar els pesos i retornar de la funcio
			break;
		}
		else	// Ja tenim h calculada. Anem a calcular c, s i w. Tornem a recorrer els vertex del triangle
		{	
			for (i=0; i<3; i++)
			{				
				// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles
				next = (i+1+3) % 3;
				prev = (i-1+3) % 3;

				c[i] = ((2.0*sin(h)*sin(h-theta[i])) / (sin(theta[next])*sin(theta[prev]))) - 1.0f;

				// Calculem el signe del determinant
				float v1[3] = {u[idx*npc*3+faces[k*3+0]*3+0], u[idx*npc*3+faces[k*3+0]*3+1], u[idx*npc*3+faces[k*3+0]*3+2]};
				float v2[3] = {u[idx*npc*3+faces[k*3+1]*3+0], u[idx*npc*3+faces[k*3+1]*3+1], u[idx*npc*3+faces[k*3+1]*3+2]};
				float v3[3] = {u[idx*npc*3+faces[k*3+2]*3+0], u[idx*npc*3+faces[k*3+2]*3+1], u[idx*npc*3+faces[k*3+2]*3+2]};
				
				res = v1[0]*v2[1]*v3[2] + v1[1]*v2[2]*v3[0] + v1[2]*v2[0]*v3[1] - v3[0]*v2[1]*v1[2] - v3[1]*v2[2]*v1[0] - v3[2]*v2[0]*v1[1];
				if (res >= 0.0)
					sign = 1;
				else
					sign = -1;
					
				s[i] = sign * sqrt(1-(c[i]*c[i]));
				
				// Preguntem si x es fora del mateix pla. En aquest cas ignorem el triangle 
				if ( fabs(s[i]) <= TOL )
				{
					jump = TRUE;
					break;
				}
			}

			// Preguntem si s'ha d'ignorar el triangle
			if (jump)
			{
				jump = FALSE;
				continue;
			}
			
			// Anem a calcular els pesos wi
			for (i=0; i<3; i++)
			{
				// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles
				next = (i+1+3) % 3;
				prev = (i-1+3) % 3;

				// Computem els pesos
				wi = (theta[i] - (c[next]*theta[prev]) - (c[prev]*theta[next])) / (dist[idx*npc+faces[k*3+i]] * sin(theta[next]) * s[prev]);
				w[idx*npc+faces[k*3+i]] += wi; 			
				sum += wi;
			}
		}
	}
	
	// Normalitzem els pesos
	for(i=0; i<npc; i++)
		w[idx*npc+i] /= sum;
}
