#include "parallel.h"
#include "utils.h"

const char *kernel_src = 
"//#pragma OPENCL EXTENSION cl_khr_fp64 : enable \n"
"\n"
"#define TRUE 1\n"
"#define FALSE 0\n"
"\n"
"__kernel void recover_point (__global float *mvcoords, __global float *cage, unsigned int cage_npts, __global float *out)\n"
"{\n"
"	float out_x = 0.0f, out_y = 0.0f, out_z = 0.0f;\n"
"	unsigned int i;\n"
"\n"
"	// Get global index in mvcoords array\n"
"	unsigned int idx = get_global_id(0);\n"
"	unsigned int k = idx*cage_npts;\n"
"\n"
"	// Calculate recovered point p\n"
"	for (i=0; i<cage_npts; i++)\n"
"	{\n"
"		out_x += mvcoords[k+i] * cage[i*3+0];\n"
"		out_y += mvcoords[k+i] * cage[i*3+1];\n"
"		out_z += mvcoords[k+i] * cage[i*3+2];\n"
"	}\n"
"\n"
"	// Store point p in out 	 \n"
"	out[idx*3+0] = out_x;\n"
"	out[idx*3+1] = out_y;\n"
"	out[idx*3+2] = out_z;\n"
"}\n"
"\n"
"__kernel void compute_coordinates (unsigned int cage_npts, __global float *cage_points, unsigned int nfaces, __global unsigned int *faces, __global float *x, __global float *w, __global float *dist, __global float *u)\n"
"{\n"
"	int sign;\n"
"	unsigned int nt=nfaces, npc=cage_npts, i, k, next, prev, p, inext, iprev, jump=FALSE;\n"
"	float l[3], theta[3], h, c[3], s[3], res, sum=0.0f, wi, ll, TOL=0.00001f;\n"
"	unsigned int idx = get_global_id(0);\n"
"	\n"
"	// Per cada vertex de la caixa calculem el modul de la distancia entre cada punt de la caixa i el punt x\n"
"	for (i=0; i<npc; i++)\n"
"	{\n"
"		dist[idx*npc+i] = sqrt(pow(cage_points[i*3+0]-x[idx*3+0], 2.0f) + pow(cage_points[i*3+1]-x[idx*3+1], 2.0f) + pow(cage_points[i*3+2]-x[idx*3+2], 2.0f));	// Calcul Modul	\n"
"\n"
"		if (dist[idx*npc+i] < TOL)	// Si el punt es molt a prop de la caixa assignem 1 al pes corresponent i la resta 0's \n"
"		{		\n"
"			w[idx*npc+i] = 1.0f;\n"
"			return;\n"
"		}\n"
"		\n"
"		// Normalitzem el vector distancia pel seu modul\n"
"		u[idx*npc*3+i*3+0] = (cage_points[i*3+0]-x[idx*3+0]) / dist[idx*npc+i];\n"
"		u[idx*npc*3+i*3+1] = (cage_points[i*3+1]-x[idx*3+1]) / dist[idx*npc+i];\n"
"		u[idx*npc*3+i*3+2] = (cage_points[i*3+2]-x[idx*3+2]) / dist[idx*npc+i];\n"
"	}\n"
"\n"
"	for (k=0; k<nt; k++)	// Per cada triangle de la caixa\n"
"	{		\n"
"		h = 0.0f;\n"
"		for (i=0; i<3; i++)	//Per cada vertex de cada triangle\n"
"		{		 	\n"
"			// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles\n"
"			next = (i+1+3) % 3;\n"
"			prev = (i-1+3) % 3;\n"
"\n"
"			// Calculem el modul de la distancia entre el punt seguent menys l'anterior\n"
"			ll = 0.0f;\n"
"			for (p=0; p<3; p++)\n"
"			{\n"
"				inext = faces[k*3+next];\n"
"				iprev = faces[k*3+prev];\n"
"				ll += pow(u[idx*npc*3+inext*3+p] - u[idx*npc*3+iprev*3+p], 2.0f);\n"
"			}\n"
"\n"
"			l[i] = sqrt(ll);\n"
"\n"
"			// Calculem l'angle theta\n"
"			theta[i] = 2.0f * asin((l[i]/2.0f));\n"
"			\n"
"			//Calculem h\n"
"			h += theta[i] / 2.0f; \n"
"		}\n"
"		\n"
"		// Preguntem si el punt x es sobre el triangle\n"
"		if ((M_PI - h) < TOL)\n"
"		{			\n"
"			// Netejem els pesos\n"
"			for (i=0; i<npc; i++)\n"
"				w[idx*npc+i] = 0.0f;\n"
"\n"
"			// Netejem el sumatori\n"
"			sum = 0.0f;\n"
"\n"
"			// Anem a calcular els pesos wi\n"
"			for (i=0; i<3; i++)\n"
"			{\n"
"				// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles\n"
"				next = (i+1+3) % 3;\n"
"				prev = (i-1+3) % 3;\n"
"\n"
"				// Computem els pesos\n"
"				wi = sin(theta[i]) * l[prev] * l[next];\n"
"				w[idx*npc+faces[k*3+i]] += wi; 			\n"
"				sum += wi;\n"
"			}\n"
"		\n"
"			// Sortim de l'evaluacio dels triangles, anem a normalitzar els pesos i retornar de la funcio\n"
"			break;\n"
"		}\n"
"		else	// Ja tenim h calculada. Anem a calcular c, s i w. Tornem a recorrer els vertex del triangle\n"
"		{	\n"
"			for (i=0; i<3; i++)\n"
"			{				\n"
"				// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles\n"
"				next = (i+1+3) % 3;\n"
"				prev = (i-1+3) % 3;\n"
"\n"
"				c[i] = ((2.0*sin(h)*sin(h-theta[i])) / (sin(theta[next])*sin(theta[prev]))) - 1.0f;\n"
"\n"
"				// Calculem el signe del determinant\n"
"				float v1[3] = {u[idx*npc*3+faces[k*3+0]*3+0], u[idx*npc*3+faces[k*3+0]*3+1], u[idx*npc*3+faces[k*3+0]*3+2]};\n"
"				float v2[3] = {u[idx*npc*3+faces[k*3+1]*3+0], u[idx*npc*3+faces[k*3+1]*3+1], u[idx*npc*3+faces[k*3+1]*3+2]};\n"
"				float v3[3] = {u[idx*npc*3+faces[k*3+2]*3+0], u[idx*npc*3+faces[k*3+2]*3+1], u[idx*npc*3+faces[k*3+2]*3+2]};\n"
"				\n"
"				res = v1[0]*v2[1]*v3[2] + v1[1]*v2[2]*v3[0] + v1[2]*v2[0]*v3[1] - v3[0]*v2[1]*v1[2] - v3[1]*v2[2]*v1[0] - v3[2]*v2[0]*v1[1];\n"
"				if (res >= 0.0)\n"
"					sign = 1;\n"
"				else\n"
"					sign = -1;\n"
"					\n"
"				s[i] = sign * sqrt(1-(c[i]*c[i]));\n"
"				\n"
"				// Preguntem si x es fora del mateix pla. En aquest cas ignorem el triangle \n"
"				if ( fabs(s[i]) <= TOL )\n"
"				{\n"
"					jump = TRUE;\n"
"					break;\n"
"				}\n"
"			}\n"
"\n"
"			// Preguntem si s'ha d'ignorar el triangle\n"
"			if (jump)\n"
"			{\n"
"				jump = FALSE;\n"
"				continue;\n"
"			}\n"
"			\n"
"			// Anem a calcular els pesos wi\n"
"			for (i=0; i<3; i++)\n"
"			{\n"
"				// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles\n"
"				next = (i+1+3) % 3;\n"
"				prev = (i-1+3) % 3;\n"
"\n"
"				// Computem els pesos\n"
"				wi = (theta[i] - (c[next]*theta[prev]) - (c[prev]*theta[next])) / (dist[idx*npc+faces[k*3+i]] * sin(theta[next]) * s[prev]);\n"
"				w[idx*npc+faces[k*3+i]] += wi;\n"
"				sum += wi;\n"
"			}\n"
"		}\n"
"	}\n"
"	\n"
"	// Normalitzem els pesos\n"
"	for(i=0; i<npc; i++)\n"
"		w[idx*npc+i] /= sum;\n"
"};\n";

void chk (cl_int status, const char* cmd) {

	if(status != CL_SUCCESS) {
		printf("%s failed (%d)\n", cmd, status);
		exit(-1);
	}
}

void init_opencl_resources (struct cl_data *data) 
{
	// Use this to check the output of each API call
    	cl_int status;  
     
	//-----------------------------------------------------
	// STEP 1: Discover and initialize the platforms
	//-----------------------------------------------------

	// Use clGetPlatformIDs() to retrieve the number of 
	// platforms
	status = clGetPlatformIDs(0, NULL, &(data->numPlatforms));
	chk(status, "clGetPlatformIDs");

	// Allocate enough space for each platform
	data->platforms = (cl_platform_id*)malloc(data->numPlatforms*sizeof(cl_platform_id));

	// Fill in platforms with clGetPlatformIDs()
	status = clGetPlatformIDs(data->numPlatforms, data->platforms, NULL);
	chk(status, "clGetPlatformIDs");
	
	//-----------------------------------------------------
	// STEP 2: Discover and initialize the devices
	//----------------------------------------------------- 

	// Use clGetDeviceIDs() to retrieve the number of 
	// devices present
	status = clGetDeviceIDs(data->platforms[0], CL_DEVICE_TYPE_GPU, 0, NULL, &(data->numDevices));
	chk(status, "clGetDeviceIDs");

	// Allocate enough space for each device
	data->devices = (cl_device_id*)malloc(data->numDevices*sizeof(cl_device_id));

	// Fill in devices with clGetDeviceIDs()
	status = clGetDeviceIDs(data->platforms[0], CL_DEVICE_TYPE_GPU, data->numDevices, data->devices, NULL);
	chk(status, "clGetDeviceIDs");

	//-----------------------------------------------------
	// STEP 3: Create a context
	//----------------------------------------------------- 

	// Create a context using clCreateContext() and 
	// associate it with the devices
	data->context = clCreateContext(NULL, data->numDevices, data->devices, NULL, NULL, &status);
	chk(status, "clCreateContext");

	//-----------------------------------------------------
	// STEP 4: Create a command queue
	//----------------------------------------------------- 

	// Create a command queue using clCreateCommandQueue(),
	// and associate it with the device you want to execute 
	// on
	data->cmdQueue = clCreateCommandQueue(data->context, data->devices[0], 0, &status);
	chk(status, "clCreateCommandQueue");

	//-----------------------------------------------------
	// STEP 7: Create and compile the program
	//----------------------------------------------------- 

	// Create a program using clCreateProgramWithSource()
	//const char *programSource = readSource("../kernel.cl");
	data->program = clCreateProgramWithSource(data->context, 1, (const char**)&kernel_src, NULL, &status);
	chk(status, "clCreateProgramWithSource");

	// Build (compile) the program for the devices with
	// clBuildProgram()
	status = clBuildProgram(data->program, data->numDevices, data->devices, NULL, NULL, NULL);
	chk(status, "clBuildProgram");
}

void compute_parallel_coordinates (struct cl_data *clData, Cage *cage, unsigned int fig_npts, float *figure_points, float *mvcoords)
{
	// Use this to check the output of each API call
    	cl_int status;

	//-----------------------------------------------------
	// STEP 5: Create device buffers
	//----------------------------------------------------- 	
	
	// Use clCreateBuffer() to create a buffer object (d_C) 
	// with enough space to hold the output data
	clData->bufferMV = clCreateBuffer(clData->context, CL_MEM_READ_WRITE, fig_npts * cage->npts * sizeof(float), NULL, &status);
	chk(status, "clCreateBuffer for output");
	
	// Use clCreateBuffer() to create a buffer object (d_B)
	// that will contain the data from the host array B
	clData->bufferCage = clCreateBuffer(clData->context, CL_MEM_READ_ONLY, sizeof(float) * cage->npts * 3, NULL, &status);
	chk(status, "clCreateBuffer for cage");
	
	// Use clCreateBuffer() to create a buffer object (d_B)
	// that will contain the data from the host array B
	clData->bufferFigure = clCreateBuffer(clData->context, CL_MEM_READ_WRITE, sizeof(float) * fig_npts * 3, NULL, &status);
	chk(status, "clCreateBuffer for figure");
	
	// Use clCreateBuffer() to create a buffer object (d_B)
	// that will contain the data from the host array B
	clData->bufferFaces = clCreateBuffer(clData->context, CL_MEM_READ_ONLY, sizeof(unsigned int) * cage->nfaces * 3, NULL, &status);
	chk(status, "clCreateBuffer for faces");
	
	// Use clCreateBuffer() to create a buffer object (d_B)
	// that will contain the data from the host array B
	clData->bufferDist = clCreateBuffer(clData->context, CL_MEM_READ_WRITE, sizeof(float) * cage->npts * fig_npts, NULL, &status);
	chk(status, "clCreateBuffer for dist");	
	
	// Use clCreateBuffer() to create a buffer object (d_B)
	// that will contain the data from the host array B	
	clData->bufferU = clCreateBuffer(clData->context, CL_MEM_READ_WRITE, sizeof(float) * cage->npts * 3 * fig_npts, NULL, &status);
	chk(status, "clCreateBuffer for u");	
	
	//-----------------------------------------------------
	// STEP 6: Write host data to device buffers
	//----------------------------------------------------- 

	// Use clEnqueueWriteBuffer() to write input array B to 
	// the device buffer bufferB
	status = clEnqueueWriteBuffer(clData->cmdQueue, clData->bufferMV, CL_FALSE, 0, sizeof(float) * cage->npts * fig_npts, mvcoords, 0, NULL, NULL);
	chk(status, "clEnqueueWriteBuffer for cage");
	
	// Use clEnqueueWriteBuffer() to write input array B to 
	// the device buffer bufferB
	status = clEnqueueWriteBuffer(clData->cmdQueue, clData->bufferCage, CL_FALSE, 0, sizeof(float) * cage->npts * 3, cage->points, 0, NULL, NULL);
	chk(status, "clEnqueueWriteBuffer for cage");
	
	// Use clEnqueueWriteBuffer() to write input array B to 
	// the device buffer bufferB
	status = clEnqueueWriteBuffer(clData->cmdQueue, clData->bufferFigure, CL_FALSE, 0, sizeof(float) * fig_npts * 3, figure_points, 0, NULL, NULL);
	chk(status, "clEnqueueWriteBuffer for figure");
	
	// Use clEnqueueWriteBuffer() to write input array B to 
	// the device buffer bufferB
	status = clEnqueueWriteBuffer(clData->cmdQueue, clData->bufferFaces, CL_FALSE, 0, sizeof(unsigned int) * cage->nfaces * 3, cage->faces, 0, NULL, NULL);
	chk(status, "clEnqueueWriteBuffer for faces");
	
	// Use clEnqueueWriteBuffer() to write input array B to 
	// the device buffer bufferB
	float *dist = (float *) calloc(cage->npts * fig_npts, sizeof(float));
	status = clEnqueueWriteBuffer(clData->cmdQueue, clData->bufferDist, CL_FALSE, 0, sizeof(float) * cage->npts * fig_npts, dist, 0, NULL, NULL);
	chk(status, "clEnqueueWriteBuffer for dist");
	
	// Use clEnqueueWriteBuffer() to write input array B to 
	// the device buffer bufferB
	float *u = (float *) calloc(cage->npts * 3 * fig_npts, sizeof(float));
	status = clEnqueueWriteBuffer(clData->cmdQueue, clData->bufferU, CL_FALSE, 0, sizeof(float) * cage->npts * 3 * fig_npts, u, 0, NULL, NULL);
	chk(status, "clEnqueueWriteBuffer for u");

	//-----------------------------------------------------
	// STEP 8: Create the kernel
	//----------------------------------------------------- 

	// Use clCreateKernel() to create a kernel from the 
	// vector addition function (named "vecadd")
	clData->kernel = clCreateKernel(clData->program, "compute_coordinates", &status);
	chk(status, "clCreateKernel");

	//-----------------------------------------------------
	// STEP 9: Set the kernel arguments
	//----------------------------------------------------- 

	// Associate the input and output buffers with the 
	// kernel 
	// using clSetKernelArg()
	status  = clSetKernelArg(clData->kernel, 0, sizeof(unsigned int), &cage->npts);
	status |= clSetKernelArg(clData->kernel, 1, sizeof(cl_mem), &(clData->bufferCage));
	status |= clSetKernelArg(clData->kernel, 2, sizeof(unsigned int), &cage->nfaces);
	status |= clSetKernelArg(clData->kernel, 3, sizeof(cl_mem), &(clData->bufferFaces));
	status |= clSetKernelArg(clData->kernel, 4, sizeof(cl_mem), &(clData->bufferFigure));
	status |= clSetKernelArg(clData->kernel, 5, sizeof(cl_mem), &(clData->bufferMV));
	status |= clSetKernelArg(clData->kernel, 6, sizeof(cl_mem), &(clData->bufferDist));
	status |= clSetKernelArg(clData->kernel, 7, sizeof(cl_mem), &(clData->bufferU));
	chk(status, "clSetKernelArg");
   
	//-----------------------------------------------------
	// STEP 10: Configure the work-item structure
	//----------------------------------------------------- 

	// There are 'n figure points' work-items 
	clData->globalWorkSize[0] = fig_npts;

	//-----------------------------------------------------
	// STEP 11: Enqueue the kernel for execution
	//----------------------------------------------------- 

	// Execute the kernel by using 
	// clEnqueueNDRangeKernel().
	// 'globalWorkSize' is the 1D dimension of the 
	// work-items
	status = clEnqueueNDRangeKernel(clData->cmdQueue, clData->kernel, 1, NULL, clData->globalWorkSize, NULL, 0, NULL, NULL);
	chk(status, "clEnqueueNDRangeKernel");

	//-----------------------------------------------------
    	// STEP 12: Read the output buffer back to the host
    	//----------------------------------------------------- 
    
    	// Use clEnqueueReadBuffer() to read the OpenCL output  
    	// buffer (bufferC) 
    	// to the host output array (C)
    	status = clEnqueueReadBuffer(clData->cmdQueue, clData->bufferMV, CL_TRUE, 0, sizeof(float) * fig_npts * cage->npts, mvcoords, 0, NULL, NULL);
	chk(status, "clEnqueueReadBuffer output");

	// Free data
	clReleaseMemObject(clData->bufferFaces);
	clReleaseMemObject(clData->bufferU);
	clReleaseMemObject(clData->bufferDist);
	clReleaseKernel(clData->kernel);
	free(dist);
	free(u);
}

void recover_parallel_point (struct cl_data *clData, float *cage_points, unsigned int cage_npts, unsigned int fig_npts, float *figure_points)
{
	//Use this to check the output of each API call
	cl_int status;

	// Use clEnqueueWriteBuffer() to write input array B to 
	// the device buffer bufferB
	status = clEnqueueWriteBuffer(clData->cmdQueue, clData->bufferCage, CL_FALSE, 0, sizeof(float) * cage_npts * 3, cage_points, 0, NULL, NULL);
	chk(status, "clEnqueueWriteBuffer for cage");		

	//-----------------------------------------------------
	// STEP 11: Enqueue the kernel for execution
	//----------------------------------------------------- 

	// Execute the kernel by using 
	// clEnqueueNDRangeKernel().
	// 'globalWorkSize' is the 1D dimension of the 
	// work-items
	status = clEnqueueNDRangeKernel(clData->cmdQueue, clData->kernel, 1, NULL, clData->globalWorkSize, NULL, 0, NULL, NULL);
	chk(status, "clEnqueueNDRangeKernel");

	//-----------------------------------------------------
	// STEP 12: Read the output buffer back to the host
	//----------------------------------------------------- 

	// Use clEnqueueReadBuffer() to read the OpenCL output  
	// buffer (bufferC) 
	// to the host output array (C)
	status = clEnqueueReadBuffer(clData->cmdQueue, clData->bufferFigure, CL_TRUE, 0, sizeof(float) * fig_npts * 3, figure_points, 0, NULL, NULL);
	chk(status, "clEnqueueReadBuffer");
}
