/*! \file utils.cpp
 *  \brief This file contains util functions 
 *  \author Albert Busqué
*/

#include "utils.h"
#include "const.h"

/*! 
 * \brief Compute 3-dim determinants
 *
 * This function is called internally by ::compute_coordinates
 * 
 * \param[in] v1 First row vector of matrix
 * \param[in] v2 Second row vector of matrix
 * \param[in] v3 Third row vector of matrix
 *
 * \return The result of the calculation of the determinant
*/
double determinant3x3 (double *v1, double *v2, double *v3)
{
	return v1[0]*v2[1]*v3[2] + v1[1]*v2[2]*v3[0] + v1[2]*v2[0]*v3[1] - v3[0]*v2[1]*v1[2] - v3[1]*v2[2]*v1[0] - v3[2]*v2[0]*v1[1];
}

/*! 
 * \brief Check if two vectors are equals
 *
 * This function is called internally by ::OnMouseMove
 * 
 * \param[in] a Row vector of matrix
 * \param[in] b Row vector of matrix
 * \param[in] dim Lenght of vectors
 *
 * \return The result of the comparaison. If are equals return TRUE; FALSE otherwise.
*/
unsigned int vector_equals (double *a, double *b, unsigned int dim)
{
	unsigned int i;

	for (i=0; i<dim; i++)
		if (a[i] != b[i])
			return FALSE;
	return TRUE;
}

/*! 
 * \brief Check if file exists
 *
 * This function is called internally by ::visualize
 * 
 * \param[in] filename 
 *
 * \return The result of the comparaison. If exists return TRUE; FALSE otherwise.
*/
unsigned int file_exists(char *file)
{
	FILE *fp;

	fp = fopen (file, "r");

	if (!fp)
	{
		return FALSE;
	}
	fclose(fp);
	return TRUE;
}

char *get_extension (char* filename)
{
	char* ext;

	if (!filename)
		return NULL;

	filename++;
	filename = strchr(filename, '.');

	ext = get_extension (filename);
	if (!ext)
		return filename;
	return ext;
}

// This function reads in a text file and stores it as a char pointer
/*const char* readSource (char* kernelPath) {

	int status;
	FILE *fp;
	char *source;
	long int size;

	printf("Program file is: %s\n", kernelPath);

	fp = fopen(kernelPath, "rb");
	if(!fp) {
		printf("Could not open kernel file\n");
		exit(-1);
	}
	status = fseek(fp, 0, SEEK_END);
	if(status != 0) {
		printf("Error seeking to end of file\n");
		exit(-1);
	}
	size = ftell(fp);
	if(size < 0) {
		printf("Error getting file position\n");
		exit(-1);
	}

	rewind(fp);

	source = (char *)malloc(size + 1);

	int i;
	for (i = 0; i < size+1; i++) {
		source[i]='\0';
	}

	if(source == NULL) {
		printf("Error allocating space for the kernel source\n");
		exit(-1);
	}

	fread(source, 1, size, fp);
	source[size] = '\0';

	return source;
}*/
